/*
 * bg_analyzer.c
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <bg.h>

///////////////////////////////////////////////////////////////////////////////
static bg_visitor_vtbl_t analyzer_vtbl;

int bg_analyzer_create(bg_visitor_t *vis)
{
  /////////////////////////////////////////////////////////////////////////////
  vis->vtbl=&analyzer_vtbl;
  vis->depth=0;

  /////////////////////////////////////////////////////////////////////////////
  return 0;
}

static void bg_analyzer_destroy(bg_visitor_t *vis FFUNUSED)
{
}

///////////////////////////////////////////////////////////////////////////////
static int bg_analyzer_dispatch_file(bg_visitor_t *vis FFUNUSED,
    bg_tree_t *tree FFUNUSED)
{
  // nothing to do.
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
static int bg_analyzer_print_prefix(bg_visitor_t *vis, bg_tree_t *tree)
{
  FILE *f=tree->param->result.f;
#if 0 // [
  bg_print_vmt_t *stdout_vtbl=&bg_print_classic_vmt;
#else // ] [
  bg_print_vmt_t *stdout_vtbl=tree->param->print.vtbl;
#endif // ]
  int stdout_prefix=BG_TREE_TYPE_TRACK==tree->vtbl->type||stdout_vtbl->infix;

  if (stdout_prefix) {
    ///////////////////////////////////////////////////////////////////////////
    // the header of a track is always mirrored to the shell/console
    // regardless wether prefix or infix.
    if (stdout_vtbl->head(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing head");
      goto e_print_head_stdout;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (tree->param->print.vtbl->infix) {
    if (stdout!=f&&tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
      DMESSAGE("printing head");
      goto e_print_head_file_infix;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
e_print_head_file_infix:
e_print_head_stdout:
  return -1;
}

static int bg_analyzer_print_postfix(bg_visitor_t *vis, bg_tree_t *tree)
{
  FILE *f=tree->param->result.f;
#if 0 // [
  bg_print_vmt_t *stdout_vtbl=&bg_print_classic_vmt;
#else // ] [
  bg_print_vmt_t *stdout_vtbl=tree->param->print.vtbl;
#endif // ]
  int stdout_prefix=BG_TREE_TYPE_TRACK==tree->vtbl->type||stdout_vtbl->infix;

  if (!stdout_prefix) {
    if (stdout_vtbl->head(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing head");
      goto e_print_head_stdout;
    }
  }

  if (!tree->param->print.vtbl->infix) {
    if (stdout!=f&&tree->param->print.vtbl->head(tree,vis->depth,f)<0) {
      DMESSAGE("printing head");
      goto e_print_head_file_postfix;
    }
  }

  if (tree->param->print.vtbl->tail(tree,vis->depth,f)<0) {
    DMESSAGE("printing tail");
    goto e_print_tail_file;
  }

  fflush(f);

  if (stdout!=f) {
    if (tree->param->print.vtbl->tail(tree,vis->depth,stdout)<0) {
      DMESSAGE("printing tail");
      goto e_print_tail_stdout;
    }

    fflush(stdout);
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
e_print_tail_stdout:
e_print_tail_file:
e_print_head_file_postfix:
e_print_head_stdout:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
static int bg_analyzer_dispatch_track(bg_visitor_t *vis, bg_tree_t *tree)
{
  int err=-1;
#if defined (FF_INPUT_ERROR_FORMAT) // [
  bg_track_t *track=&tree->track;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  ++vis->depth; // [

#if defined (BG_PARAM_SEQUENTIAL) // [
  /////////////////////////////////////////////////////////////////////////////
  if (bg_tree_stats_create(tree)<0) {
    DMESSAGE("creating stats");
    goto e_stats;
  }
#endif // ]

#if defined (FF_INPUT_ERROR_FORMAT) // [
  /////////////////////////////////////////////////////////////////////////////
  if (ff_input_open_analyzer(&track->input)<0) {
    DMESSAGE("re-opening input");
    goto e_input;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (ff_analyzer_create(&track->analyzer,&track->input)<0) {
    DMESSAGE("creating analyzer");
    goto e_analyzer;
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_print_prefix(vis,tree)<0) {
    DMESSAGE("printing prefix");
    goto e_print_prefix;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (ff_analyzer_loop(&tree->track.analyzer)<0) {
    DMESSAGE("decoding");
    goto e_decode;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_print_postfix(vis,tree)<0) {
    DMESSAGE("printing postfix");
    goto e_print_postfix;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=0;
//cleanup:
e_print_postfix:
e_decode:
e_print_prefix:
#if defined (FF_INPUT_ERROR_FORMAT) // [
  ff_analyzer_destroy(&track->analyzer,0);
e_analyzer:
  ff_input_close(&track->input);
e_input:
#endif // ]
#if defined (BG_PARAM_SEQUENTIAL) // [
  if (err)
    bg_tree_stats_destroy(tree);
e_stats:
#endif // ]
  --vis->depth; // ]
  return err;
}

static int bg_analyzer_dispatch_album(bg_visitor_t *vis, bg_tree_t *tree)
{
  int err=-1;
  bg_album_t *album=&tree->album;
#if defined (BG_PARAM_SEQUENTIAL) // [
  bg_param_t *param=tree->param;
  bg_tree_t *next=album->first,*cur=next;
  bg_visitor_t muxer;
#else // ] [
  bg_tree_t *cur;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  ++vis->depth; // [

#if defined (BG_PARAM_SEQUENTIAL) // [
  if (!param->sequential) {
    ///////////////////////////////////////////////////////////////////////////
    if (BG_TREE_TYPE_ROOT==tree->vtbl->type)
      err=bg_root_annotation_create(tree);
    else
      err=bg_album_annotation_create(tree);

    if (err<0) {
      DMESSAGE("annotating");
      goto e_annotate;
    }

    err=-1;
  }
#endif // ]

#if defined (BG_PARAM_SEQUENTIAL) // [
  /////////////////////////////////////////////////////////////////////////////
  memset(&muxer,0,sizeof muxer);

  /////////////////////////////////////////////////////////////////////////////
  if (bg_tree_stats_create(tree)<0) {
    DMESSAGE("creating stats");
    goto e_stats;
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_print_prefix(vis,tree)<0) {
    DMESSAGE("printing prefix");
    goto e_print_prefix;
  }

  /////////////////////////////////////////////////////////////////////////////
  for (cur=album->first;cur;cur=cur->next) {
    if (cur->vtbl->accept(cur,vis)<0) {
      FFVMESSAGE(FFL("analyzing \"%s\""),cur->source.path);
      goto e_child;
    }

    if (bg_tree_merge(tree,cur)<0) {
      DMESSAGE("merging");
      goto e_merge;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_print_postfix(vis,tree)<0) {
    DMESSAGE("printing postfix");
    goto e_print_postfix;
  }

#if defined (BG_PARAM_SEQUENTIAL) // [
  if (!param->sequential&&album->nleaves) {
    ///////////////////////////////////////////////////////////////////////////
    if (param->output.dirname||param->overwrite) {
      if (bg_muxer_create(&muxer)<0) {
        DMESSAGE("creating muxer");
        goto e_muxer;
      }

      if (!param->suppress.progress) {
        if (&bg_print_xml_vmt==param->print.vtbl)
          fputs("<!-- ",stdout);

        fputs("begin remuxing ...\n",stdout);
        fflush(stdout);
      }

      if (tree->vtbl->accept(tree,&muxer)<0) {
        DMESSAGE("muxing");
        goto e_muxerer_accept;
      }

      if (!param->suppress.progress) {
        fputs("end remuxing.",stdout);

        if (&bg_print_xml_vmt==param->print.vtbl)
          fputs(" -->",stdout);

        fputc('\n',stdout);
        fflush(stdout);
      }
    }
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  err=0;
//cleanup:
#if defined (BG_PARAM_SEQUENTIAL) // [
e_muxerer_accept:
  if (muxer.vtbl)
    muxer.vtbl->destroy(&muxer);
e_muxer:
  if (!param->sequential&&album->nleaves) {
    if (BG_TREE_TYPE_ROOT==tree->vtbl->type)
      bg_root_annotation_destroy(tree);
    else
      bg_album_annotation_destroy(tree);
  }
#endif // ]
e_print_postfix:
e_merge:
e_child:
e_print_prefix:
#if defined (BG_PARAM_SEQUENTIAL) // [
  if (err)
    bg_tree_stats_destroy(tree);
e_stats:
  if (!param->sequential) {
    if (BG_TREE_TYPE_ROOT==tree->vtbl->type)
      bg_root_annotation_destroy(tree);
    else
      bg_album_annotation_destroy(tree);
  }
e_annotate:
  if (!param->sequential)
    bg_album_destroy(tree);
#endif // ]
  --vis->depth; // ]
  return err;
}

static int bg_analyzer_dispatch_root(bg_visitor_t *vis, bg_tree_t *tree)
{
  int err=-1;

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_dispatch_album(vis,tree)<0) {
    DMESSAGE("analyzing");
    goto e_album;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=0;
e_album:
  return err;
}

static bg_visitor_vtbl_t analyzer_vtbl={
#if defined (PBU_DEBUG) // [
  .id="analyzer",
#endif // ]
  .destroy=bg_analyzer_destroy,
  .dispatch_file=bg_analyzer_dispatch_file,
  .dispatch_track=bg_analyzer_dispatch_track,
  .dispatch_album=bg_analyzer_dispatch_album,
  .dispatch_root=bg_analyzer_dispatch_root,
};
